# For Frame
from Crypto.Random import get_random_bytes

# For Handshake
import base64
import hashlib

from .httpParser import *

__all__ = ['Frame', 'Handshake', 'RFCError']

OPCODES = {
    0: 'Continuation',
    1: 'Text',
    2: 'Binary',
    3: 'Reserved_non_control_1',
    4: 'Reserved_non_control_2',
    5: 'Reserved_non_control_3',
    6: 'Reserved_non_control_4',
    7: 'Reserved_non_control_5',
    8: 'Close',
    9: 'Ping',
    10: 'Pong',
    11: 'Reserved_control_1',
    12: 'Reserved_control_2',
    13: 'Reserved_control_3',
    14: 'Reserved_control_4',
    15: 'Reserved_control_5',
}

STATUS_CODES = {
    1000: 'Normal_close',
    1001: 'Endpoint_going_away',
    1002: 'Protocol_error',
    1003: 'Data_accept_error',
    1004: 'Reserved_status_code_1004',
    1005: 'Reserved_status_code_1005',
    1006: 'Reserved_status_code_1006',
    1007: 'Data_type_error',
    1008: 'Policy_violation',
    1009: 'Message_too_big',
    1010: 'Client_extension_error',
    1011: 'Unexpected_condition',
    1012: 'TLS_handshake_error',
}


class Frame:
    """Contains methods to pack and unpack a websocket frame
    """

    # ----------------- Status codes ------------------------
    STATUS_CODE_NORMAL_CLOSE = 1000
    STATUS_CODE_ENDPOINT_GOING_AWAY = 1001
    STATUS_CODE_PROTOCOL_ERROR = 1002
    STATUS_CODE_DATA_ACCEPT_ERROR = 1003
    STATUS_CODE_RESERVED_STATUS_CODE_1004 = 1004
    STATUS_CODE_RESERVED_STATUS_CODE_1005 = 1005
    STATUS_CODE_RESERVED_STATUS_CODE_1006 = 1006
    STATUS_CODE_DATA_TYPE_ERROR = 1007
    STATUS_CODE_POLICY_VIOLATION = 1008
    STATUS_CODE_MESSAGE_TOO_BIG = 1009
    STATUS_CODE_CLIENT_EXTENSION_ERROR = 1010
    STATUS_CODE_UNEXPECTED_CONDITION = 1011
    STATUS_CODE_TLS_HANDSHAKE_ERROR = 1012

    @staticmethod
    def is_status_code(status_code):
        if 1000 <= status_code <= 1012:
            return True
        return False

    # --------------- Opcodes ----------------------
    OPCODE_CONTINUATION = 0
    OPCODE_TEXT = 1
    OPCODE_BINARY = 2
    OPCODE_RESERVED_NON_CONTROL_1 = 3
    OPCODE_RESERVED_NON_CONTROL_2 = 4
    OPCODE_RESERVED_NON_CONTROL_3 = 5
    OPCODE_RESERVED_NON_CONTROL_4 = 6
    OPCODE_RESERVED_NON_CONTROL_5 = 7
    OPCODE_CLOSE = 8
    OPCODE_PING = 9
    OPCODE_PONG = 10
    OPCODE_RESERVED_CONTROL_1 = 11
    OPCODE_RESERVED_CONTROL_2 = 12
    OPCODE_RESERVED_CONTROL_3 = 13
    OPCODE_RESERVED_CONTROL_4 = 14
    OPCODE_RESERVED_CONTROL_5 = 15

    @staticmethod
    def is_opcode(opcode):
        return 0 <= opcode <= 15

    # ---------------------------------------------------

    def __init__(self, fin=1, rsv=(0, 0, 0), opcode=None, mask=0, masking_key=None,
                 payload_length=None, payload=b''):
        self.fin = fin
        self.rsv = rsv
        self.opcode = opcode
        self.mask = mask
        self.masking_key = masking_key
        self.payload_length = payload_length
        self.payload = payload
        self.extended_payload_byte_length = None
        self.payload_missing = None

    def unpack_header(self, header):
        """
        :param header: The websocket frame header (2 bytes)
        :return: Returns a **Frame** object
        """
        self.fin = header[0] >> 7
        self.rsv = (header[0] >> 6 & 0b00000001,
                    header[0] >> 5 & 0b00000001,
                    header[0] >> 4 & 0b00000001)
        self.opcode = header[0] & 0b00001111
        if not self.is_opcode(self.opcode):
            raise RFCError(
                RFCError.ERROR_INVALID_OPCODE,
                'Invalid opcode ({})'.format(self.opcode)
            )

        self.mask = header[1] >> 7
        self.payload_length = header[1] & 0b01111111
        self.payload_missing = self.payload_length - len(self.payload)
        if self.payload_length < 126:
            self.extended_payload_byte_length = 0
        elif self.payload_length == 126:
            self.extended_payload_byte_length = 2
        elif self.payload_length == 127:
            self.extended_payload_byte_length = 8

        return self

    def pack(self):
        if self.opcode is None:
            raise RFCError(
                RFCError.ERROR_INVALID_OPCODE,
                'Opcode not set'
            )
        elif not self.is_opcode(self.opcode):
            raise RFCError(
                RFCError.ERROR_INVALID_OPCODE,
                'Invalid opcode ({})'.format(self.opcode)
            )

        frame_head = bytearray(2)
        frame_head[0] = self.fin << 7
        frame_head[0] |= self.rsv[0] << 6
        frame_head[0] |= self.rsv[1] << 5
        frame_head[0] |= self.rsv[2] << 4
        frame_head[0] |= self.opcode

        frame_head[1] = self.mask << 7

        payload_length = len(self.payload)
        if payload_length < 126:
            payload_length_ext = bytearray(0)
            frame_head[1] |= payload_length
        elif payload_length < 65536:
            payload_length_ext = payload_length.to_bytes(2, 'big')
            frame_head[1] |= 126
        elif payload_length < 18446744073709551616:
            payload_length_ext = payload_length.to_bytes(8, 'big')
            frame_head[1] |= 127
        else:
            raise RFCError(
                RFCError.ERROR_PAYLOAD_TOO_LARGE,
                'Payload too large ({})'.format(payload_length)
            )

        payload = self.payload
        if self.mask:
            self.masking_key = get_random_bytes(4)
            payload = self.masking_algorithm(self.payload, self.masking_key)
        else:
            self.masking_key = b''

        return bytes(frame_head + payload_length_ext + self.masking_key + payload)

    def add_payload(self, payload):
        """ Adds bytes to the payload variable.

        If mask = 1 unmasks payload before adding

        :param payload: Bytes to add to the payload
        :return: Returns unmasked payload
        """

        if self.mask:
            payload = self.masking_algorithm(payload, self.masking_key, len(self.payload))

        self.payload += payload
        self.payload_missing = self.payload_length - len(self.payload)

        return payload

    @staticmethod
    def masking_algorithm(data, key, start=0):
        result = bytearray(len(data))
        for i in range(0, len(data)):
            result[i] = data[i] ^ key[(i + start) % 4]
        return bytes(result)

    def __str__(self):
        payload = self.payload[0:10]
        payload = str(payload)
        if len(payload) < len(self.payload):

            payload += '...'

        return '<{}Frame opcode={}, payload_length={}, payload={}, mask={}, key={}>'.format(
            OPCODES[self.opcode], self.opcode, self.payload_length, payload, self.mask, self.masking_key
        )

    def close_frame(self, status_code, payload=b''):
        """
        :param status_code: The close frame status code
        :param payload: Additional information to send with the close frame
        :return: returns packed close frame
        """
        self.opcode = self.OPCODE_CLOSE
        if not self.is_status_code(status_code):
            raise RFCError(
                RFCError.ERROR_INVALID_STATUS_CODE,
                'Invalid status code ({})'.format(status_code)
            )

        self.payload = int.to_bytes(status_code, 2, 'big') + payload
        return self.pack()

    def set_payload_length(self, data):
        """
        Args:
             data: Bytes representing the payload length
        """
        self.payload_length = int.from_bytes(data, 'big')
        self.payload_missing = self.payload_length - len(self.payload)


class Handshake:
    GUID = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
    CLIENT_KEY_HEADER = b'Sec-WebSocket-Key'
    SERVER_KEY_HEADER = b'Sec-WebSocket-Accept'
    BASE_SERVER_HANDSHAKE = HttpResponse(
        status_code=b'101 Switching Protocols',
        headers={
            b'Upgrade': b'websocket',
            b'Connection': b'Upgrade',
            # b'Sec-WebSocket-Protocol': b'chat, superchat',
        }
    )
    BASE_CLIENT_HANDSHAKE = HttpRequest(
        method=b'GET',
        headers={
            b'Upgrade': b'websocket',
            b'Connection': b'Upgrade',
            b'Sec-WebSocket-Protocol': b'chat',
            b'Sec-WebSocket-Version': b'13',
        }
    )

    @staticmethod
    def get_server_key(client_key):
        return base64.b64encode(hashlib.sha1(client_key+Handshake.GUID).digest())

    @staticmethod
    def get_client_key():
        random_bytes = get_random_bytes(16)
        random_bytes = base64.b64encode(random_bytes)
        return random_bytes

    def pack_server_handshake(self, client_handshake=None, client_key=None, additional_headers=None):
        if client_handshake:
            try:
                request = HttpRequest().unpack(client_handshake)
            except HttpError as err:
                raise RFCError(
                    RFCError.ERROR_INVALID_HANDSHAKE,
                    'Invalid client handshake ({})'.format(err)
                )

            # if b'Host' in request.headers:
            #     self.BASE_SERVER_HANDSHAKE.headers[b'Host'] = request.headers[b'Host']

            try:
                client_key = request.headers[self.CLIENT_KEY_HEADER]
            except KeyError:
                raise RFCError(
                    RFCError.ERROR_INVALID_HANDSHAKE,
                    'Invalid client handshake, no key header found'
                )
        elif client_key is None:
            raise ValueError("Both client_handshake and client_key can't be None type")

        if additional_headers:
            self.BASE_SERVER_HANDSHAKE.headers.update(additional_headers)

        server_key = self.get_server_key(client_key)
        self.BASE_SERVER_HANDSHAKE.headers[self.SERVER_KEY_HEADER] = server_key

        return self.BASE_SERVER_HANDSHAKE.pack()

    def pack_client_handshake(self, client_key, uri=b'/', additional_headers=None):
        """
         Returns:
             Client handshake in bytes
        """
        self.BASE_CLIENT_HANDSHAKE.headers[self.CLIENT_KEY_HEADER] = client_key
        self.BASE_CLIENT_HANDSHAKE.uri = uri
        if additional_headers:
            self.BASE_CLIENT_HANDSHAKE.headers.update(additional_headers)
        return self.BASE_CLIENT_HANDSHAKE.pack()

# Error codes


class RFCError(OSError):
    ERROR_INVALID_OPCODE = 1
    ERROR_INVALID_HANDSHAKE = 2
    ERROR_PAYLOAD_TOO_LARGE = 3
    ERROR_MISSING_BYTES = 4
    ERROR_NO_HANDSHAKE = 5
    ERROR_INVALID_STATUS_CODE = 6



"""
Contains methods for parsing an http request (No shit!)
"""

__all__ = [
    'HttpRequest',
    'HttpResponse',
    'HttpError'
    ]


REQUEST_METHODS = [b'GET', b'POST', b'PUT', b'PATCH', b'DELETE']


class Http:
    def __init__(self,
                 headers=None,
                 body=b''):

        if headers:
            self.headers = headers
        else:
            self.headers = {}
        self.body = body

    def pack_headers(self):
        data = b''
        for header in self.headers:
            data += header + b': ' + self.headers[header] + b'\r\n'

        return data

    def unpack_header(self, header):
        """ Unpacks a header line and adds it to the header dict
        ex. b'Upgrade: websocket\r\n' will make an entry in the header dict
        self.headers[b'Upgrade'] = b'websocket'
        Args:
            header: A header line (in bytes) from an http request/response
        """
        key_value = header.split(b':', 1)
        if len(key_value) != 2:
            # This line doesn't contain a b':'
            raise HttpError(
                HttpError.ERROR_INVALID_HEADER_LINE,
                'Invalid header line ({})'.format(header)
            )

        key_value[0] = key_value[0].strip()
        key_value[1] = key_value[1].strip()
        self.headers[key_value[0]] = key_value[1]

    def unpack_headers_and_body(self, headers_and_body):
        """
         Args:
             headers_and_body: All lines (in a list) except the first in an http request or response
        """
        i = 1
        for line in headers_and_body:
            i += 1
            line = line.strip()
            if line == b'':
                # End of headers
                break
            self.unpack_header(line)

        self.body = b'\r\n'.join(headers_and_body[i:])

    def pack_headers_and_body(self):
        data = b''
        for header in self.headers:
            data += header + b': ' + self.headers[header] + b'\r\n'

        data += b'\r\n' + self.body

        return data


class HttpRequest(Http):

    def __init__(self,
                 method=b'GET',
                 uri=b'/',
                 http_version=b'HTTP/1.1',
                 headers=None,
                 body=b''
                 ):
        self.method = method
        self.uri = uri
        self.http_version = http_version

        Http.__init__(self, headers, body)

    def unpack(self, request):
        lines = request.splitlines()
        if len(lines) == 0:
            raise HttpError(
                HttpError.ERROR_EMPTY_HTTP,
                'Request is emtpy'
            )

        header = lines[0].split()
        if len(header) != 3:
            raise HttpError(
                HttpError.ERROR_INVALID_REQUEST,
                'Invalid request line ({})'.format(lines[0])
            )

        if header[0] not in REQUEST_METHODS:
            raise HttpError(
                HttpError.ERROR_INVALID_REQUEST,
                'Invalid request method ({})'.format(header[0])
            )

        self.method = header[0]
        self.uri = header[1]
        self.http_version = header[2]

        self.unpack_headers_and_body(lines[1:])
        return self

    def pack(self):
        if None in [self.method, self.uri, self.http_version]:
            raise HttpError(
                HttpError.ERROR_INVALID_REQUEST,
                "Can't pack while method, uri or http_version i still None type"
            )

        request = b' '.join([self.method, self.uri, self.http_version]) + b'\r\n'
        request += self.pack_headers_and_body()
        return request

    def __str__(self):
        return '<HttpRequest, method={}, uri={}, version={}>'.format(
            self.method, self.uri, self.http_version
        )


class HttpResponse(Http):
    def __init__(self,
                 http_version=b'HTTP/1.1',
                 status_code=b'200 OK',
                 headers=None,
                 body=b''
                 ):

        Http.__init__(self, headers, body)
        self.status_code = status_code
        self.http_version = http_version

    def unpack(self, response):
        lines = response.splitlines()
        if len(lines) == 0:
            raise HttpError(
                HttpError.ERROR_EMPTY_HTTP,
                'Response is emtpy'
            )

        header = lines[0].split()
        if len(header) < 2:
            raise HttpError(
                HttpError.ERROR_INVALID_RESPONSE,
                'Response line is missing a status code'
            )
        self.http_version = header[0]
        self.status_code = b' '.join(header[1:])
        self.unpack_headers_and_body(lines[1:])
        return self

    def pack(self):
        if None in [self.http_version, self.status_code]:
            raise HttpError(
                HttpError.ERROR_INVALID_RESPONSE,
                "Can't pack while http_version or status_code i still None type"
            )

        request = b' '.join([self.http_version, self.status_code]) + b'\r\n'
        request += self.pack_headers_and_body()
        return request

    def __str__(self):
        return '<HttpResponse, version={}, status={}>'.format(
            self.http_version, self.status_code
        )


class HttpError(OSError):
    ERROR_INVALID_HEADER_LINE = 1
    ERROR_EMPTY_HTTP = 3
    ERROR_INVALID_REQUEST = 5
    ERROR_INVALID_RESPONSE = 6



#!/bin/env python3
# from distutils.core import setup
from setuptools import setup
from wswrapper import __version__
import sys

modulename = 'wswrapper'

def readme():
    with open('README.rst') as f:
        return f.read()

print('Current version: ', __version__)
version = __version__.split('.')


if sys.argv[-1] == 'minor':
    version[2] = str(int(version[2]) + 1)
    del sys.argv[-1]
elif sys.argv[-1] == 'major':
    version[1] = str(int(version[1]) + 1)
    version[2] = '0'
    del sys.argv[-1]
elif sys.argv[-1] == 'huge':
    version[0] = str(int(version[0]) + 1)
    version[1] = '0'
    version[2] = '0'
    del sys.argv[-1]


version = '.'.join(version)
with open(modulename + '/version', 'w') as f:
    f.write(version)


setup(
    name=modulename,
    packages=[modulename],  # this must be the same as the name above
    version=version,
    include_package_data=True,
    license='MIT',
    description='Wrapps a socket.socket object to handle the RFC6455 websocket protocol',
    long_description=readme(),
    author='Christoffer Zakrisson',
    author_email='christoffer_zakrisson@hotmail.com',
    url='https://github.com/Zaeb0s/ws-wrapper',
    keywords=['wrapper', 'websocket', 'socket', 'RFC6455', 'wrap'],
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Programming Language :: Python :: 3.5',
                 'Operating System :: POSIX :: Linux',
                 'License :: OSI Approved :: MIT License'],
    install_requires=['pycrypto']
)


print('Installed version: ' + version)



from .RFC6455 import *
from .httpParser import *

import socket
import errno

__all__ = [
    'Wrapper',
    'WSError'
]

class Wrapper:
    STATE_CONNECTING = 0
    STATE_OPEN = 1
    STATE_CLOSING = 2
    STATE_CLOSED = 3

    STATES = {
        STATE_CONNECTING: 'Connecting',
        STATE_OPEN: 'Open',
        STATE_CLOSING: 'Closing',
        STATE_CLOSED: 'Closed'
    }

    def __init__(self,
                 sock,
                 do_handshake_on_connect=True):

        self.sock = sock
        self.do_handshake_on_connect = do_handshake_on_connect
        self.state = self.STATE_CONNECTING
        self.current_frame = None
        self.close_frame_sent = False
        self.close_frame_recd = False
        self.close_payload = None
        self.is_client = None

        self.handshake_headers = {}

    def do_handshake(self, uri=b'/', headers=None):
        """ Does handshake as client if self.is_client else does handshake from server side
         When handshake is complete sets the websocket to open state
         Args:
             uri: The uri of the handshake http request. Only significant if doing handshake from client side.
             headers: A dict containing additional headers of http request/response
        """

        if self.is_client:
            self.do_handshake_as_client(uri, headers)
        else:
            self.do_handshake_as_server(headers)

        self.state = self.STATE_OPEN

    def do_handshake_as_client(self, uri=b'/', headers=None):
        """ Does the websocket handshake from the client side
        """
        key = Handshake.get_client_key()
        request = Handshake().pack_client_handshake(
            client_key=key,
            uri=uri,
            additional_headers=headers
        )
        self.sock.send(request)
        response = self.sock.recv(1024)
        try:
            response = HttpResponse().unpack(response)
        except HttpError as err:
            raise WSError(
                WSError.ERROR_INVALID_HANDSHAKE_RESPONSE,
                'Received invalid handshake response from server ({})'.format(err)
            )

        try:
            server_key = response.headers[Handshake.SERVER_KEY_HEADER]
        except KeyError:
            raise WSError(
                WSError.ERROR_INVALID_HANDSHAKE_RESPONSE,
                'Did not find server key header in handshake response'
            )

        if server_key != Handshake.get_server_key(key):
            raise WSError(
                WSError.ERROR_INVALID_HANDSHAKE_RESPONSE,
                'Received invalid server response key during WS handshake'
            )

    def do_handshake_as_server(self, headers=None):
        """ Does the websocket handshake from the server side
        """
        request = self.sock.recv(1024)
        try:
            request = HttpRequest().unpack(request)
        except HttpError as err:
            raise WSError(
                WSError.ERROR_INVALID_HANDSHAKE_REQUEST,
                'Received invalid handshake request from client ({})'.format(err)
            )

        try:
            client_key = request.headers[Handshake.CLIENT_KEY_HEADER]
        except KeyError:
            raise WSError(
                WSError.ERROR_INVALID_HANDSHAKE_REQUEST,
                'Did not find client key header in handshake request'
            )

        try:
            response = Handshake().pack_server_handshake(
                client_key=client_key,
                additional_headers=headers,
            )

        except RFCError as err:
            raise WSError(
                WSError.ERROR_INVALID_HANDSHAKE_REQUEST,
                'Received invalid client handshake request ({})'.format(err)
            )

        self.sock.send(response)

    def accept(self):
        conn, addr = self.sock.accept()

        client = Wrapper(conn)
        client.is_client = False
        if self.do_handshake_on_connect:
            client.do_handshake(headers=self.handshake_headers)

        return client, addr

    def recv_frame_info(self):
        """ Receives the information about the payload of the current frame in the buffer.
        The frame header, extended payload length and masking key

        Raises:
            WSError: With errno ERROR_FRAME_INFO_FAILED if any of the information could not be obtained
        """
        frame_header = self.sock.recv(2)
        if len(frame_header) != 2:
            raise WSError(
                WSError.ERROR_FRAME_INFO_FAILED,
                'Failed to receive frame header'
            )

        try:
            frame = Frame().unpack_header(frame_header)
        except RFCError as err:
            # Invalid frame header
            raise WSError(
                WSError.ERROR_FRAME_INFO_FAILED,
                'Failed to process frame header ({})'.format(err.args[1])
            )

        if frame.extended_payload_byte_length != 0:
            extended_payload_bytes = self.sock.recv(frame.extended_payload_byte_length)
            if len(extended_payload_bytes) != frame.extended_payload_byte_length:
                raise WSError(
                    WSError.ERROR_FRAME_INFO_FAILED,
                    'Failed to receive extended payload length'
                )
            frame.set_payload_length(extended_payload_bytes)

        if frame.mask:
            frame.masking_key = self.sock.recv(4)
            if len(frame.masking_key) != 4:
                raise WSError(
                    WSError.ERROR_FRAME_INFO_FAILED,
                    'Failed to receive masking key'
                )

        self.current_frame = frame

    def recv(self, buffersize, flags=0):
        # Reset current frame if all of the payload has been received
        if self.current_frame and self.current_frame.payload_missing == 0:
            self.current_frame = None

        # TODO: Implement flags = socket.MSG_PEEK
        if self.current_frame is None:
            try:
                self.recv_frame_info()
            except OSError as err:
                if err.args[0] == errno.EWOULDBLOCK:
                    raise WSError(
                        WSError.ERROR_FRAME_INFO_FAILED,
                        'Frame info missing in buffer'
                    )
                else:
                    raise

        frame = self.current_frame

        if frame.opcode == Frame.OPCODE_CLOSE:
            self.close_frame_recd = True
            self.close_payload = self.recv_payload(128)
            return b''

        elif frame.opcode in [Frame.OPCODE_TEXT, Frame.OPCODE_BINARY]:
            return self.recv_payload(buffersize)

        raise WSError(
            WSError.ERROR_FRAME_INFO_FAILED,
            'Unimplemented opcode'
        )

    def send(self, data, opcode=Frame.OPCODE_TEXT):
        if self.state not in [self.STATE_CLOSING, self.STATE_OPEN]:
            raise WSError(
                WSError.ERROR_STATE_ERROR,
                "Can't send websocket frame while websocket is in {} state".format(self.STATES[self.state])
            )

        frame = Frame(
            opcode=opcode,
            payload=data,
            mask=self.is_client
        )

        self.sock.send(frame.pack())

    def recv_payload(self, buffersize):
        """ Receives payload of the current frame, unmasks if the payload is masked before returning
        Args:
            buffersize: Number of bytes to receive, if buffersize is larger than the payload left for
             the current frame will only receive the rest of the current frame
        Returns:
             Unmasked payload bytes

        """
        frame = self.current_frame

        buffersize = min(frame.payload_missing, buffersize)

        payload = self.sock.recv(buffersize)
        payload = frame.add_payload(payload)

        return payload

    def shutdown(self, flag=socket.SHUT_RDWR, status_code=Frame.STATUS_CODE_ENDPOINT_GOING_AWAY):
        self.state = self.STATE_CLOSING
        if self.state != self.STATE_CONNECTING:
            if not self.close_frame_sent:
                if self.close_payload:
                    close_frame = Frame(
                        opcode=Frame.OPCODE_CLOSE,
                        payload=self.close_payload
                    ).pack()
                else:
                    close_frame = Frame().close_frame(status_code=status_code)

                self.sock.send(close_frame)

        self.sock.shutdown(flag)

    def close(self):
        self.sock.close()
        self.state = self.STATE_CLOSED

    def setblocking(self, flag):
        self.sock.setblocking(flag)

    def setsockopt(self, level, option, value):
        self.sock.setsockopt(level, option, value)

    def connect(self, address):
        self.is_client = True
        self.sock.connect(address)
        if self.do_handshake_on_connect:
            self.do_handshake(headers=self.handshake_headers)

    def listen(self, backlog):
        self.sock.listen(backlog)

    def bind(self, address):
        self.is_client = False
        self.sock.bind(address)


class WSError(OSError):
    ERROR_INVALID_HANDSHAKE_RESPONSE = 1
    ERROR_INVALID_HANDSHAKE_REQUEST = 2
    ERROR_FRAME_INFO_FAILED = 3
    ERROR_STATE_ERROR = 4
